from __future__ import annotations

import time
from datetime import UTC, datetime
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from limits import RateLimitItem
    from starlette.datastructures import MutableHeaders

    from .config import RateLimitConfig
    from .datastructures import RateLimitInfo


def make_legacy_headers(info: RateLimitInfo) -> dict[str, str]:
    return {
        "X-RateLimit-Limit": str(info.limit),
        "X-RateLimit-Remaining": str(info.remaining),
        "X-RateLimit-Reset": str(info.reset_time),
        # Provide a Date header to deal with time desynchronization
        # between the client and server.
        "Date": datetime.now(UTC).strftime("%a, %d %b %Y %H:%M:%S %Z"),
    }


def make_draft_06_headers(
    info: RateLimitInfo, window: int, reset_secs: int
) -> dict[str, str]:
    return {
        "RateLimit-Policy": f"{info.limit};w={window}",
        "RateLimit-Limit": str(info.limit),
        "RateLimit-Remaining": str(info.remaining),
        "RateLimit-Reset": str(reset_secs),
    }


def make_draft_07_headers(
    info: RateLimitInfo, window: int, reset_secs: int
) -> dict[str, str]:
    return {
        "RateLimit-Policy": f"{info.limit};w={window}",
        "RateLimit": f"limit={info.limit}, remaining={info.remaining}, reset={reset_secs}",  # noqa: E501
    }


def make_headers(
    config: RateLimitConfig, limit: RateLimitItem, info: RateLimitInfo
) -> dict[str, str]:
    headers: dict[str, str] = {}
    reset_secs = info.reset_time - int(time.time())

    if config.legacy_headers:
        headers.update(make_legacy_headers(info))

    if config.headers == "draft-06":
        headers.update(make_draft_06_headers(info, limit.get_expiry(), reset_secs))
    elif config.headers == "draft-07":
        headers.update(make_draft_07_headers(info, limit.get_expiry(), reset_secs))

    if (config.legacy_headers or config.headers is not None) and info.remaining == 0:
        headers["Retry-After"] = str(reset_secs)

    return headers


def clear_headers(headers: MutableHeaders):
    del headers["X-RateLimit-Limit"]
    del headers["X-RateLimit-Remaining"]
    del headers["X-RateLimit-Reset"]
    del headers["Date"]
    del headers["RateLimit-Policy"]
    del headers["RateLimit"]
    del headers["RateLimit-Limit"]
    del headers["RateLimit-Remaining"]
    del headers["RateLimit-Reset"]



__title__ = "fourtwonine"
__author__ = "beerpsi <beerpsi@duck.com>"
__license__ = "MIT"
__copyright__ = "Copyright 2024-present Liminova"
__version__ = "0.1.0"

from .config import RateLimitConfig
from .datastructures import RateLimitInfo
from .errors import RateLimitExceeded
from .fastapi import RateLimitDep
from .limiter import get_config, initialize_config
from .middleware import RateLimitMiddleware

__all__ = (
    "RateLimitConfig",
    "RateLimitInfo",
    "RateLimitExceeded",
    "RateLimitDep",
    "get_config",
    "initialize_config",
    "RateLimitMiddleware",
)



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated, NoReturn, cast

import limits
from limits import RateLimitItem
from starlette.applications import Starlette
from starlette.datastructures import Headers, MutableHeaders
from starlette.requests import Request
from typing_extensions import Doc

from ._core import check_limits
from ._utils import ensure_config_exists, maybe_await
from .errors import RateLimitExceeded

if TYPE_CHECKING:
    from starlette.types import ASGIApp, Message, Receive, Scope, Send

    from .config import RateLimitConfig

__all__ = ("RateLimitMiddleware",)


class RateLimitMiddleware:
    """ASGI rate limiting middleware for Starlette."""

    def __init__(
        self,
        app: ASGIApp,
        limit_value: Annotated[
            str | RateLimitItem,
            Doc(
                """
                A rate limit passed as a `limits.RateLimitItem` or a string
                (e.g. `10/second`).
                """
            ),
        ],
        bucket: Annotated[str, Doc("The rate limit bucket to use.")],
        config: Annotated[
            RateLimitConfig | None,
            Doc(
                """
                The rate limit configuration to use. Uses the global configuration
                attached to your app if not specified.
                """
            ),
        ] = None,
    ) -> None:
        self.app = app
        self.bucket = bucket
        self.config = config

        if isinstance(limit_value, str):
            self.limit_value = limits.parse(limit_value)
        else:
            self.limit_value = limit_value

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        self.config = ensure_config_exists(self.config, cast(Starlette, scope["app"]))
        request = Request(scope, receive, send)

        try:
            headers = await check_limits(
                request, self.config, self.limit_value, self.bucket
            )
        except RateLimitExceeded as e:
            response = await maybe_await(self.config.response_func(request, e))

            await response(scope, receive, send)
            return

        if headers is None:
            return

        await RateLimitResponder(self.app, headers)(scope, receive, send)


class RateLimitResponder:
    def __init__(self, app: ASGIApp, headers: dict[str, str]) -> None:
        self.app = app
        self.send: Send = unattached_send
        self.headers = headers

        self.started: bool = False
        self.rate_limit_headers_set: bool = False
        self.initial_message: Message = {}

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        self.send = send
        await self.app(scope, receive, self.send_wrapper)

    async def send_wrapper(self, message: Message):
        if message["type"] == "http.response.start":
            self.initial_message = message

            raw_headers: list[tuple[bytes, bytes]] = self.initial_message["headers"]  # pyright: ignore[reportRedeclaration]
            headers = Headers(raw=raw_headers)

            if "X-RateLimit-Limit" in headers or "RateLimit-Policy" in headers:
                self.rate_limit_headers_set = True

        # Middleware in Starlette goes from bottom to top. If there are already
        # rate-limit headers set (by a more local rate limiter), don't set
        # rate limit headers any more.
        elif message["type"] == "http.response.body" and self.rate_limit_headers_set:
            if not self.started:
                self.started = True
                await self.send(self.initial_message)

            await self.send(message)
        elif message["type"] == "http.response.body":
            if not self.started:
                self.started = True

                raw_headers: list[tuple[bytes, bytes]] = self.initial_message["headers"]
                headers = MutableHeaders(raw=raw_headers)
                headers.update(self.headers)

                await self.send(self.initial_message)

            await self.send(message)


async def unattached_send(_message: Message) -> NoReturn:
    raise RuntimeError("send awaitable not set")



from __future__ import annotations

from typing import TYPE_CHECKING

from starlette.exceptions import HTTPException

if TYPE_CHECKING:
    from .config import RateLimitConfig


class RateLimitExceeded(HTTPException):
    """Exception thrown when the rate limit has exceeded."""

    def __init__(
        self,
        config: RateLimitConfig,
        detail: str | None = None,
        headers: dict[str, str] | None = None,
    ):
        super().__init__(config.status_code, detail, headers)



from dataclasses import dataclass

from limits import RateLimitItem
from limits.aio.strategies import RateLimiter

from .config import RateLimitConfig


@dataclass
class RateLimitInfo:
    """
    The rate limit information for each client, attached in the
    :cls:`starlette.requests.Request.state` object.
    """

    limiter: RateLimiter
    item: RateLimitItem
    config: RateLimitConfig
    key: str

    remaining: int
    reset_time: int

    @property
    def limit(self):
        return self.item.amount

    @property
    def used(self):
        return self.item.amount - self.remaining



from typing import cast

from starlette.applications import Starlette

from .config import RateLimitConfig


def initialize_config(app: Starlette, config: RateLimitConfig):
    app.state.__fourtwonine_config = config  # noqa: SLF001


def get_config(app: Starlette):
    return cast(RateLimitConfig, app.state.__fourtwonine_config)  # noqa: SLF001



# pyright: reportImportCycles=false
from __future__ import annotations

from collections.abc import Awaitable, Callable
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal

from starlette.responses import JSONResponse
from starlette.status import HTTP_429_TOO_MANY_REQUESTS

if TYPE_CHECKING:
    from limits.aio.strategies import RateLimiter
    from starlette.requests import Request
    from starlette.responses import Response

    from .errors import RateLimitExceeded

StandardHeaderType = Literal["draft-06", "draft-07"]
KeyFuncT = Callable[["Request", str], Awaitable[str] | str]
SkipFuncT = Callable[["Request"], Awaitable[bool] | bool]
DetailFuncT = Callable[["Request"], Awaitable[str | None] | str | None]
ResponseFuncT = Callable[
    ["Request", "RateLimitExceeded"], "Awaitable[Response] | Response"
]


@dataclass(kw_only=True)
class RateLimitConfig:
    """Rate limiting configuration."""

    limiter: RateLimiter
    """
    The :cls:`limits.aio.strategies.RateLimiter` instance used to throttle requests.
    """

    pass_on_limiter_error: bool = False
    """
    Whether to block or allow requests in the case of a limiter error. If this is
    set to `False` (default), all requests are blocked until the limiter no longer errors
    (fail closed). Set this to `True` to allow requests through without rate limiting.
    """

    status_code: int = HTTP_429_TOO_MANY_REQUESTS
    """The response code used for rate limited requests."""

    legacy_headers: bool = False
    """
    Whether to send legacy rate limit headers for the limit (`X-RateLimit-Limit`), current
    usage (`X-RateLimit-Remaining`) and reset time (`X-RateLimit-Reset`) on all responses.

    If a request is blocked, `Retry-After` will also be sent.
    """

    headers: None | StandardHeaderType = "draft-07"
    """
    Whether to send rate limit headers conforming to the
    [RateLimit header fields for HTTP standardization draft](https://github.com/ietf-wg-httpapi/ratelimit-headers)
    adopted by the IETF.

    If set to `draft-06`, the headers `RateLimit-Policy`, `RateLimit-Limit`,
    `RateLimit-Remaining` and `RateLimit-Reset` are sent on all responses, matching the
    [sixth draft](https://datatracker.ietf.org/doc/html/draft-ietf-httpapi-ratelimit-headers-06)
    of the specification.

    If set to `draft-07`, the headers `RateLimit` and `RateLimit-Policy` are sent,
    matching the [seventh draft](https://datatracker.ietf.org/doc/html/draft-ietf-httpapi-ratelimit-headers-07)
    of the specification.

    If ratelimit headers are sent and a request is blocked, `Retry-After` will also
    be sent.
    """

    key_func: KeyFuncT = lambda r, b: f"{r.client.host if r.client else 'anonymous'}:{b}"
    """
    Function to retrieve custom identifiers for clients. Should be a sync/async
    function that accepts :cls:`starlette.requests.Request` and the bucket name,
    and returns a string.

    By default, the IP address of the client is used:

    ```python
    lambda r, b: lambda r, b: f"{r.client.host if r.client else 'anonymous'}:{b}"
    ```
    """

    skip_func: SkipFuncT = lambda _: False
    """
    Function to determine whether this request counts towards a client's quota.
    Should be a sync/async function that accepts :cls:`starlette.requests.Request`
    and returns a boolean.

    By default, no requests are skipped:
    ```python
    lambda _: False
    ```
    """

    error_detail_func: DetailFuncT = lambda _: "Rate limit exceeded."
    """
    Function to determine the detail attached to the rate limit exception.
    Should be a sync/async function that accepts :cls:`starlette.requests.Request`
    and returns a string.
    """

    response_func: ResponseFuncT = lambda _, e: JSONResponse(
        {"detail": e.detail},
        e.status_code,
        e.headers,
    )
    """
    Function to make a response when the rate limit is hit. Only used on Starlette
    middleware; the FastAPI dependency raises :cls:`fourtwonine.errors.RateLimitExceeded`
    directly.
    """

    state_attr: str = "rate_limit"
    """
    The name of the attribute on the :attr:`starlette.requests.Request.state` object
    to store the rate limit information.
    """



# Do not get tempted by nicer type annotations! Using `from __future__ import annotations`
# will break Pydantic!
from typing import Annotated, Optional, Union, cast

import limits
from limits import RateLimitItem
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from typing_extensions import Doc

from ._core import check_limits
from ._headers import clear_headers
from ._utils import ensure_config_exists
from .config import RateLimitConfig


class RateLimitDep:
    """A rate limiting dependency for FastAPI."""

    def __init__(
        self,
        limit_value: Annotated[
            Union[str, RateLimitItem],
            Doc(
                """
                A rate limit passed as a `limits.RateLimitItem` or a string
                (e.g. `10/second`).
                """
            ),
        ],
        bucket: Annotated[str, Doc("The rate limit bucket to use.")],
        config: Annotated[
            Optional[RateLimitConfig],
            Doc(
                """
                The rate limit configuration to use. Uses the global configuration
                attached to your app if not specified.
                """
            ),
        ] = None,
    ) -> None:
        if isinstance(limit_value, str):
            self.limit_value = limits.parse(limit_value)
        else:
            self.limit_value = limit_value

        self.bucket = bucket
        self.config = config

    async def __call__(self, request: Request, response: Response):
        self.config = ensure_config_exists(self.config, cast(Starlette, request.app))
        headers = await check_limits(request, self.config, self.limit_value, self.bucket)

        if headers is None:
            return

        # FastAPI dependencies are evaluated from top-to-bottom; clear all headers
        # so we can apply the innermost rate limit headers.
        clear_headers(response.headers)
        response.headers.update(headers)



from __future__ import annotations

import inspect
from typing import TYPE_CHECKING, TypeVar, cast

from .config import RateLimitConfig

if TYPE_CHECKING:
    from collections.abc import Awaitable

    from starlette.applications import Starlette


T = TypeVar("T")


async def maybe_await(value: Awaitable[T] | T) -> T:
    if inspect.isawaitable(value):
        return await value

    return value


def ensure_config_exists(config: RateLimitConfig | None, app: Starlette):
    if config is not None:
        return config

    try:
        config = cast(RateLimitConfig, app.state.__fourtwonine_config)  # noqa: SLF001
    except AttributeError as e:
        msg = (
            "This dependency does not specify a config, but a global config was "
            "not initialized. Use `fourtwonine.initialize` to attach a global "
            "rate limit configuration to your app, or specify a config."
        )
        raise RuntimeError(msg) from e

    return config



from limits.aio.storage import MemoryStorage
from limits.aio.strategies import MovingWindowRateLimiter
from starlette.applications import Starlette
from starlette.middleware import Middleware
from starlette.requests import Request
from starlette.responses import PlainTextResponse
from starlette.routing import Route

import fourtwonine
from fourtwonine import RateLimitConfig, RateLimitMiddleware


async def rate_limited_endpoint(_request: Request):
    return PlainTextResponse("hit me 11 times in a minute!")


async def hello(_request: Request):
    return PlainTextResponse("hello world!")


app = Starlette(
    # Apply a global rate limit...
    middleware=[
        Middleware(RateLimitMiddleware, limit_value="20/minute", bucket="global"),
    ],
    routes=[
        Route(
            "/rate-limited",
            rate_limited_endpoint,
            # ...and then a specific rate limit for an endpoint.
            middleware=[
                Middleware(
                    RateLimitMiddleware, limit_value="10/minute", bucket="rate-limited"
                ),
            ],
        ),
        Route("/hello", hello),
    ],
)

# Attach a global configuration to the app, which will be used for
# all middlewares that don't specify their own configuration.
fourtwonine.initialize_config(
    app,
    RateLimitConfig(
        limiter=MovingWindowRateLimiter(MemoryStorage()),
    ),
)



from __future__ import annotations

from typing import TYPE_CHECKING

from limits.errors import StorageError

from ._headers import make_headers
from ._utils import maybe_await
from .datastructures import RateLimitInfo
from .errors import RateLimitExceeded

if TYPE_CHECKING:
    from limits import RateLimitItem
    from starlette.requests import Request

    from .config import RateLimitConfig


async def check_limits(
    request: Request, config: RateLimitConfig, limit: RateLimitItem, bucket: str
):
    if await maybe_await(config.skip_func(request)):
        return None

    limiter = config.limiter
    key = await maybe_await(config.key_func(request, bucket))

    try:
        acquired = await limiter.hit(limit, key)
        reset, remaining = await limiter.get_window_stats(limit, key)
    except Exception as e:
        if config.pass_on_limiter_error and isinstance(
            e, (StorageError, limiter.storage.base_exceptions)
        ):
            return None

        raise

    info = RateLimitInfo(
        limiter,
        limit,
        config,
        key,
        remaining,
        reset,
    )
    headers = make_headers(config, limit, info)

    setattr(request.state, config.state_attr, info)

    if not acquired:
        detail = await maybe_await(config.error_detail_func(request))

        raise RateLimitExceeded(config, detail, headers)

    return headers



from fastapi import Depends, FastAPI
from limits.aio.storage import MemoryStorage
from limits.aio.strategies import MovingWindowRateLimiter

import fourtwonine
from fourtwonine.config import RateLimitConfig
from fourtwonine.fastapi import RateLimitDep

app = FastAPI(
    # Apply a global rate limit...
    dependencies=[Depends(RateLimitDep("20/minute", "global"))]
)
fourtwonine.initialize_config(
    app,
    config=RateLimitConfig(
        limiter=MovingWindowRateLimiter(MemoryStorage()),
        skip_func=lambda r: r.url.path in {"/redoc", "/docs", "/openapi.json"},
    ),
)


@app.get(
    "/rate-limited",
    # ...and then a specific rate limit for an endpoint.
    dependencies=[Depends(RateLimitDep("10/minute", "rate-limited"))],
)
async def rate_limited():
    return {"challenge": "hit me 11 times in a minute!"}


# This endpoint will still be affected by the global rate limit.
@app.get("/hello")
async def hello_world():
    return {"message": "Hello World!"}
